import { createCallback, createSubRoot } from "@solid-primitives/rootless";
import {
  batch,
  createMemo,
  createRenderEffect,
  createSignal,
  Index,
  Match,
  onCleanup,
  onMount,
  Switch,
  type Component,
} from "solid-js";
import { render } from "solid-js/web";
import Scaffold from "~material/Scaffold";
import { isPointNotInRect } from "./dom";
import "./MediaQuickview.css";
import AppTopBar from "~material/AppTopBar";
import { IconButton } from "@suid/material";
import { Close } from "@suid/icons-material";
import { createStore, unwrap } from "solid-js/store";

function renderIsolateMediaQuickview(
  each: QuickviewMedia[],
  index: number,
  transitionFrom?: Element,
) {
  createSubRoot((disposeAll) => {
    let container: HTMLDivElement;

    createRenderEffect(() => {
      container = document.createElement("div");
      container.setAttribute("role", "presentation");
      container.classList.add("MediaQuickview__root");
      document.querySelector("body")!.appendChild(container);

      onCleanup(() => container.remove());

      const dispose = render(() => {
        return (
          <MediaQuickview
            each={each}
            defaultIndex={index}
            transitonFrom={transitionFrom}
            onClose={disposeAll}
          />
        );
      }, container);

      onCleanup(dispose);
    });
  });
}

export function createMediaQuickview() {
  return createCallback(renderIsolateMediaQuickview);
}

function ImagePage(props: {
  src: string;
  alt?: string;
  scale: number;
  offsetX: number;
  offsetY: number;
}) {
  return (
    <img
      src={props.src}
      alt={props.alt}
      style={{
        transform: `scale(${props.scale}) translateX(${-props.offsetX}px) translateY(${-props.offsetY}px)`,
      }}
    ></img>
  );
}

export type QuickviewMedia = {
  cat: "image" | "video" | "gifv" | "audio" | "unknown";
  src: string;
  alt?: string;
};

export type MediaQuickviewProps = {
  each: QuickviewMedia[];
  defaultIndex: number;
  transitonFrom?: Element;

  onClose?(): void;
};

function clamp<T extends number>(value: T, min: T, max: T) {
  return Math.max(Math.min(value, max), min);
}

const MediaQuickview: Component<MediaQuickviewProps> = (props) => {
  let root: HTMLDialogElement;

  const [lightOut, setLightOut] = createSignal(false);

  function onDialogClick(
    event: MouseEvent & { currentTarget: HTMLDialogElement },
  ) {
    event.stopPropagation();

    if (
      isPointNotInRect(
        event.currentTarget.getBoundingClientRect(),
        event.clientX,
        event.clientY,
      )
    ) {
      event.currentTarget.close();
    } else {
      setLightOut((x) => !x);
    }
  }

  const [transformations, setTransformations] = createStore(
    [] as {
      scale: number;
      /**
       * positive = the left edge move towards left
       */
      offsetX: number;
      /**
       * positive = the top edge move towards top
       */
      offsetY: number;
    }[],
  );

  const transformationGetOrSetDefault = (index: number) => {
    if (transformations.length <= index) {
      setTransformations(index, {
        scale: 1,
        offsetX: 0,
        offsetY: 0,
      });
    }
    return transformations[index];
  };

  const onWheel = (
    index: number,
    event: WheelEvent & { currentTarget: HTMLElement },
  ) => {
    // This is a de-facto standard for scaling:
    // Browsers will simulate ctrl + wheel for two point scaling gesture on trackpad.
    if (event.ctrlKey) {
      event.preventDefault();
      event.stopPropagation();

      const { clientX, clientY, currentTarget, deltaY } = event;

      const offset = -deltaY; // not reversed wheel: wheel up = scale +, wheel down = scale -
      const scaleOffset = offset / currentTarget.clientHeight;

      // Map the screen scale to the image viewport scale
      const userOriginX = clientX - currentTarget.clientLeft;
      const userOriginY = clientY - currentTarget.clientTop;
      setTransformations(index, ({ scale, offsetX, offsetY }) => {
        const nscale = scale + scaleOffset;
        return {
          offsetX: offsetX + (userOriginX * nscale - userOriginX),
          offsetY: offsetY + (userOriginY * nscale - userOriginX),
          scale: nscale,
        };
      });
    }
  };

  const [isMoving, setIsMoving] = createSignal(false);
  let movOriginX: number = 0,
    movOriginY: number = 0;

  const onMouseDown = (event: MouseEvent) => {
    event.preventDefault();
    setIsMoving(true);
    movOriginX = event.clientX;
    movOriginY = event.clientY;
  };

  const onMouseMove = (index: number, event: MouseEvent) => {
    if (!isMoving()) {
      return;
    }
    const dx = movOriginX - event.clientX;
    const dy = movOriginY - event.clientY ;

    setTransformations(index, ({ offsetX, offsetY }) => {
      return {
        offsetX: offsetX + dx,
        offsetY: offsetY + dy,
      };
    });

    movOriginX = event.clientX;
    movOriginY = event.clientY;
  };

  const onMouseUp = (event: MouseEvent) => {
    setIsMoving(false);
  };

  return (
    <dialog
      ref={(e) => {
        root = e;
        onMount(() => {
          e.showModal();
        });
      }}
      class="MediaQuickview"
      classList={{ lightout: lightOut(), moving: isMoving() }}
      onClose={props.onClose}
      onCancel={props.onClose}
      onClick={onDialogClick}
    >
      <Scaffold
        topbar={
          <AppTopBar>
            <IconButton
              color="inherit"
              onClick={(e) => root.close()}
              disableFocusRipple
            >
              <Close />
            </IconButton>
          </AppTopBar>
        }
      >
        <div
          ref={(e) => {
            onMount(() => {
              e.children.item(props.defaultIndex)!.scrollIntoView({
                behavior: "instant",
                inline: "center",
              });
            });
          }}
          class="pages"
        >
          <Index each={props.each}>
            {(item, index) => {
              return (
                <section
                  class="page"
                  onWheel={[onWheel, index]}
                  onMouseDown={onMouseDown}
                  onMouseMove={[onMouseMove, index]}
                  onMouseUp={onMouseUp}
                >
                  <Switch>
                    <Match when={item().cat === "image"}>
                      <ImagePage
                        src={item().src}
                        alt={item().alt}
                        scale={transformationGetOrSetDefault(index).scale}
                        offsetX={transformationGetOrSetDefault(index).offsetX}
                        offsetY={transformationGetOrSetDefault(index).offsetY}
                      />
                    </Match>
                  </Switch>
                </section>
              );
            }}
          </Index>
        </div>
      </Scaffold>
    </dialog>
  );
};

export default MediaQuickview;
